#include "ObjectManager.h"
#include "LevelManager.h"
#include "Utils.h"
#include "Constants.h"
#include "Player.h"
#include "sprite.h"
#define BULLET_SIZE 20
#define PLAYERBULLET_SIZE 20

#define MAX_OBJECTS 60

ObjectStruct objects[MAX_OBJECTS];
u16 numActiveObjects = 0;
u16 numKills = 0;

ObjectStruct *firstObjectOfType[MAX_OBJECT_TYPES];
ObjectStruct *lastObjectOfType[MAX_OBJECT_TYPES];

Sprite *bSprites[BULLET_SIZE];
Sprite *pSprites[PLAYERBULLET_SIZE];

u16 bActives = 0;
u16 pActives = 0;

const u8 totalKills[] = {10, 20, 30, 40};

void ObjectManager_Init()
{
    numActiveObjects = 0;
    numKills = 0;
    bActives = 0;
    pActives = 0;
    for (u8 i = 0; i < BULLET_SIZE; i++)
    {
        bSprites[i] = SPR_addSprite(&enemybullet, 0, 0, TILE_ATTR(PAL0, FALSE, FALSE, FALSE));
        SPR_setVisibility(bSprites[i], HIDDEN);
    }

    for (u8 i = 0; i < PLAYERBULLET_SIZE; i++)
    {
        pSprites[i] = SPR_addSprite(&playerbullet, 0, 0, TILE_ATTR(PAL0, FALSE, FALSE, FALSE));
        SPR_setVisibility(pSprites[i], HIDDEN);
    }
}

u16 OutsideAreaWithinHorizontalScreenSpan(ObjectStruct *object);

ObjectStruct *ObjectManager_CreateObject(u16 objectType,
                                         ObjectFunctionType updateFunction,
                                         ObjectFunctionType destroyFunction,
                                         ObjectFunctionType drawFunction,
                                         ObjectCollisionFunctionType collisionFunction)
{
    if (numActiveObjects >= MAX_OBJECTS)
    {
        return NULL;
    }

    ObjectStruct *object = objects + numActiveObjects;

    object->updateFunction = updateFunction;
    object->destroyFunction = destroyFunction;
    object->drawFunction = drawFunction;
    object->collisionFunction = collisionFunction;
    object->objectState = OBJECTSTATE_ACTIVE;
    object->objectType = objectType;

    if (object->objectType == TYPE_PLAYER_PROJECTILE)
    {
        if (pActives >= PLAYERBULLET_SIZE)
        {
            return NULL;
        }
        object->sprite = *(pSprites + pActives);
        SPR_setVisibility(object->sprite, VISIBLE);
        pActives++;
    }
    else if (object->objectType == TYPE_ENEMY_PROJECTILE)
    {
        if (bActives >= BULLET_SIZE)
        {
            return NULL;
        }
        object->sprite = *(bSprites + bActives);
        SPR_setVisibility(object->sprite, VISIBLE);
        bActives++;
    }

    numActiveObjects++;

    return object;
}

void DrawObjects(u16 objectType)
{
    if (*(firstObjectOfType + objectType) == NULL)
        return;

    ObjectStruct *object = *(firstObjectOfType + objectType);

    while (object != NULL)
    {
        object->drawFunction(object);
        object = (ObjectStruct *)object->nextVisibleObjectOfSameType;
    }
}

void ObjectManager_DrawObjects()
{
    playerObject->drawFunction(playerObject);
    DrawObjects(TYPE_ENEMY);
}

void ProcessCollisions(u16 firstType, u16 secondType)
{
    if (*(firstObjectOfType + firstType) == NULL)
        return;
    if (*(firstObjectOfType + secondType) == NULL)
        return;

    ObjectStruct *objectOfFirstType = *(firstObjectOfType + firstType);
    ObjectStruct *objectOfSecondType = NULL;

    while (objectOfFirstType != NULL)
    {
        if (objectOfFirstType->healthPoints > 0 &&
            objectOfFirstType->objectState == OBJECTSTATE_ACTIVE)
        {
            objectOfSecondType = *(firstObjectOfType + secondType);

            while (objectOfSecondType != NULL)
            {
                if (objectOfSecondType->healthPoints > 0 &&
                    objectOfSecondType->objectState == OBJECTSTATE_ACTIVE)
                {
                    fix16 width1 = intToFix16(objectOfFirstType->sprite->frame->w);
                    fix16 height1 = intToFix16(objectOfFirstType->sprite->frame->h);
                    fix16 width2 = intToFix16(objectOfSecondType->sprite->frame->w);
                    fix16 height2 = intToFix16(objectOfSecondType->sprite->frame->h);

                    if (!(objectOfFirstType->x > objectOfSecondType->x + width2 ||
                          objectOfFirstType->x + width1 < objectOfSecondType->x ||
                          objectOfFirstType->y > objectOfSecondType->y + height2 ||
                          objectOfFirstType->y + height1 < objectOfSecondType->y))
                    {
                        objectOfFirstType->collisionFunction(objectOfFirstType, objectOfSecondType);
                        objectOfSecondType->collisionFunction(objectOfSecondType, objectOfFirstType);
                        if (objectOfSecondType->objectState == OBJECTSTATE_INACTIVE)
                        {
                            AddScore(objectOfSecondType->score);
                            numKills++;

                            if (numKills > totalKills[playerObject->weapon])
                            {
                                numKills = 0;
                                PowerUpCreate(objectOfSecondType->x, objectOfSecondType->y);
                            }
                        }
                        break; //maybe not best?
                    }
                }

                objectOfSecondType = objectOfSecondType->nextVisibleObjectOfSameType;
            }
        }

        objectOfFirstType = objectOfFirstType->nextVisibleObjectOfSameType;
    }
}

void ProcessCollisionsAgainstPlayer(u16 objectType)
{
    if (*(firstObjectOfType + objectType) == NULL)
        return;

    // no need to do anything if we're dead.
    if (playerObject->objectState != OBJECTSTATE_ACTIVE)
    {
        return;
    }

    ObjectStruct *firstObject = *(firstObjectOfType + objectType);

    while (firstObject != NULL)
    {
        if (firstObject->objectState == OBJECTSTATE_ACTIVE)
        {
            if (firstObject->objectType == TYPE_ENEMY_PROJECTILE)
            {
                if (!(firstObject->x + intToFix16(2) > playerObject->x + intToFix16(11) ||
                      firstObject->x + intToFix16(5) < playerObject->x + intToFix16(5) ||
                      firstObject->y + intToFix16(1) > playerObject->y + intToFix16(6) ||
                      firstObject->y + intToFix16(4) < playerObject->y + intToFix16(2)))
                {
                    firstObject->collisionFunction(firstObject, playerObject);
                    playerObject->collisionFunction(playerObject, firstObject);
                    break; //maybe not best?
                }
            }
            else
            {
                fix16 width1 = intToFix16(firstObject->sprite->frame->w);
                fix16 height1 = intToFix16(firstObject->sprite->frame->h);
                fix16 width2 = intToFix16(playerObject->sprite->frame->w);
                fix16 height2 = intToFix16(playerObject->sprite->frame->h);

                if (!(firstObject->x > playerObject->x + width2 ||
                      firstObject->x + width1 < playerObject->x ||
                      firstObject->y > playerObject->y + height2 ||
                      firstObject->y + height1 < playerObject->y))
                {
                    firstObject->collisionFunction(firstObject, playerObject);
                    playerObject->collisionFunction(playerObject, firstObject);
                    break; //maybe not best?
                }
            }
        }

        firstObject = firstObject->nextVisibleObjectOfSameType;
    }
}

void ProcessAllCollisions()
{
    ProcessCollisions(TYPE_PLAYER_PROJECTILE, TYPE_ENEMY);
    ProcessCollisionsAgainstPlayer(TYPE_ENEMY);
    ProcessCollisionsAgainstPlayer(TYPE_ENEMY_PROJECTILE);
    ProcessCollisionsAgainstPlayer(TYPE_PICKUP);
}

void DestroyObjects()
{
    ObjectStruct *tempObject = objects;
    ObjectStruct *lastObject = (objects + numActiveObjects);

    while (tempObject != lastObject)
    {
        tempObject->objectState = OBJECTSTATE_INACTIVE;
        tempObject->destroyFunction(tempObject);
        tempObject++;
    }
    numActiveObjects = 0;
}

void UpdateObjects()
{
    ObjectStruct *tempObject = objects;
    ObjectStruct *lastObject = (objects + numActiveObjects);

    memset(firstObjectOfType, NULL, sizeof(firstObjectOfType));
    memset(lastObjectOfType, NULL, sizeof(lastObjectOfType));

    ObjectStruct **firstObjectOfTypePointer = firstObjectOfType;
    ObjectStruct **lastObjectOfTypePointer = lastObjectOfType;

    u16 activeAreaIsWithinScreenSpan = FALSE;

    while (tempObject != lastObject)
    {
        tempObject->updateFunction(tempObject);
        if (tempObject->objectType == TYPE_PICKUP)
        {
            activeAreaIsWithinScreenSpan = TRUE;
        }
        else
        {
            activeAreaIsWithinScreenSpan = !OutsideAreaWithinHorizontalScreenSpan(tempObject);
        }

        if (!activeAreaIsWithinScreenSpan)
        {
            //KLog("die");
            tempObject->objectState = OBJECTSTATE_INACTIVE;
        }

        if (tempObject->objectState == OBJECTSTATE_INACTIVE)
        {
            if (tempObject->objectType == TYPE_PLAYER_PROJECTILE)
            {
                SPR_setVisibility(tempObject->sprite, HIDDEN);
                pActives--;
                *(pSprites + pActives) = tempObject->sprite;
            }
            else if (tempObject->objectType == TYPE_ENEMY_PROJECTILE)
            {
                SPR_setVisibility(tempObject->sprite, HIDDEN);
                bActives--;
                *(bSprites + bActives) = tempObject->sprite;
            }

            tempObject->destroyFunction(tempObject); // what happens if they create any new objects here???

            numActiveObjects--;

            // set the new last object
            lastObject = (objects + numActiveObjects);

            // swap the top of the stack with the current dead object.
            // this will eliminate any holes in the stack.
            if (tempObject != lastObject)
            {
                memcpy(tempObject, lastObject, sizeof(ObjectStruct));
            }

            // we don't move the tempObject pointer because we've copied a new object
            // to where the dead one is. So we want to reprocess it.
        }
        else if (tempObject->objectState == OBJECTSTATE_ACTIVE)
        {
            tempObject->lifetime++; // loops after 65535. Maximum of 18.2 minutes.

            // update ordering of same type.
            u16 objectType = tempObject->objectType;

            if (*(firstObjectOfTypePointer + objectType) == NULL)
            {
                *(firstObjectOfTypePointer + objectType) = tempObject;
            }

            if (*(lastObjectOfTypePointer + objectType) != NULL)
            {
                (*(lastObjectOfTypePointer + objectType))->nextVisibleObjectOfSameType = tempObject;
            }

            tempObject->nextVisibleObjectOfSameType = NULL;
            *(lastObjectOfTypePointer + objectType) = tempObject;

            tempObject++;
        }
    }
}

u16 OutsideAreaWithinHorizontalScreenSpan(ObjectStruct *object)
{
    if (object->sprite == NULL)
    {
        return TRUE;
    }

    fix16 width = intToFix16(object->sprite->frame->w);
    fix16 height = intToFix16(object->sprite->frame->h);

    return object->x < 0 || object->y + height < 0 || object->x - width > intToFix16(VIRTUAL_SCREEN_WIDTH) || object->y - height > intToFix16(VIRTUAL_SCREEN_HEIGHT);
}

void ObjectManager_UpdateObjects()
{
    playerObject->updateFunction(playerObject);
    playerObject->lifetime++;

    UpdateObjects();

    ProcessAllCollisions();
}